# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.protobuf import struct_pb2  # type: ignore
from google.protobuf import timestamp_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1beta1",
    manifest={
        "ModelMonitoringStats",
        "ModelMonitoringStatsDataPoint",
        "ModelMonitoringTabularStats",
        "SearchModelMonitoringStatsFilter",
    },
)


class ModelMonitoringStats(proto.Message):
    r"""Represents the collection of statistics for a metric.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        tabular_stats (google.cloud.aiplatform_v1beta1.types.ModelMonitoringTabularStats):
            Generated tabular statistics.

            This field is a member of `oneof`_ ``stats``.
    """

    tabular_stats: "ModelMonitoringTabularStats" = proto.Field(
        proto.MESSAGE,
        number=1,
        oneof="stats",
        message="ModelMonitoringTabularStats",
    )


class ModelMonitoringStatsDataPoint(proto.Message):
    r"""Represents a single statistics data point.

    Attributes:
        current_stats (google.cloud.aiplatform_v1beta1.types.ModelMonitoringStatsDataPoint.TypedValue):
            Statistics from current dataset.
        baseline_stats (google.cloud.aiplatform_v1beta1.types.ModelMonitoringStatsDataPoint.TypedValue):
            Statistics from baseline dataset.
        threshold_value (float):
            Threshold value.
        has_anomaly (bool):
            Indicate if the statistics has anomaly.
        model_monitoring_job (str):
            Model monitoring job resource name.
        schedule (str):
            Schedule resource name.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Statistics create time.
        algorithm (str):
            Algorithm used to calculated the metrics, eg:
            jensen_shannon_divergence, l_infinity.
    """

    class TypedValue(proto.Message):
        r"""Typed value of the statistics.

        This message has `oneof`_ fields (mutually exclusive fields).
        For each oneof, at most one member field can be set at the same time.
        Setting any member of the oneof automatically clears all other
        members.

        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            double_value (float):
                Double.

                This field is a member of `oneof`_ ``value``.
            distribution_value (google.cloud.aiplatform_v1beta1.types.ModelMonitoringStatsDataPoint.TypedValue.DistributionDataValue):
                Distribution.

                This field is a member of `oneof`_ ``value``.
        """

        class DistributionDataValue(proto.Message):
            r"""Summary statistics for a population of values.

            Attributes:
                distribution (google.protobuf.struct_pb2.Value):
                    Predictive monitoring drift distribution in
                    ``tensorflow.metadata.v0.DatasetFeatureStatistics`` format.
                distribution_deviation (float):
                    Distribution distance deviation from the current dataset's
                    statistics to baseline dataset's statistics.

                    - For categorical feature, the distribution distance is
                      calculated by L-inifinity norm or Jensen–Shannon
                      divergence.
                    - For numerical feature, the distribution distance is
                      calculated by Jensen–Shannon divergence.
            """

            distribution: struct_pb2.Value = proto.Field(
                proto.MESSAGE,
                number=1,
                message=struct_pb2.Value,
            )
            distribution_deviation: float = proto.Field(
                proto.DOUBLE,
                number=2,
            )

        double_value: float = proto.Field(
            proto.DOUBLE,
            number=1,
            oneof="value",
        )
        distribution_value: (
            "ModelMonitoringStatsDataPoint.TypedValue.DistributionDataValue"
        ) = proto.Field(
            proto.MESSAGE,
            number=2,
            oneof="value",
            message="ModelMonitoringStatsDataPoint.TypedValue.DistributionDataValue",
        )

    current_stats: TypedValue = proto.Field(
        proto.MESSAGE,
        number=1,
        message=TypedValue,
    )
    baseline_stats: TypedValue = proto.Field(
        proto.MESSAGE,
        number=2,
        message=TypedValue,
    )
    threshold_value: float = proto.Field(
        proto.DOUBLE,
        number=3,
    )
    has_anomaly: bool = proto.Field(
        proto.BOOL,
        number=4,
    )
    model_monitoring_job: str = proto.Field(
        proto.STRING,
        number=5,
    )
    schedule: str = proto.Field(
        proto.STRING,
        number=6,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=7,
        message=timestamp_pb2.Timestamp,
    )
    algorithm: str = proto.Field(
        proto.STRING,
        number=8,
    )


class ModelMonitoringTabularStats(proto.Message):
    r"""A collection of data points that describes the time-varying
    values of a tabular metric.

    Attributes:
        stats_name (str):
            The stats name.
        objective_type (str):
            One of the supported monitoring objectives:
            ``raw-feature-drift`` ``prediction-output-drift``
            ``feature-attribution``
        data_points (MutableSequence[google.cloud.aiplatform_v1beta1.types.ModelMonitoringStatsDataPoint]):
            The data points of this time series. When
            listing time series, points are returned in
            reverse time order.
    """

    stats_name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    objective_type: str = proto.Field(
        proto.STRING,
        number=2,
    )
    data_points: MutableSequence["ModelMonitoringStatsDataPoint"] = proto.RepeatedField(
        proto.MESSAGE,
        number=3,
        message="ModelMonitoringStatsDataPoint",
    )


class SearchModelMonitoringStatsFilter(proto.Message):
    r"""Filter for searching ModelMonitoringStats.

    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        tabular_stats_filter (google.cloud.aiplatform_v1beta1.types.SearchModelMonitoringStatsFilter.TabularStatsFilter):
            Tabular statistics filter.

            This field is a member of `oneof`_ ``filter``.
    """

    class TabularStatsFilter(proto.Message):
        r"""Tabular statistics filter.

        Attributes:
            stats_name (str):
                If not specified, will return all the stats_names.
            objective_type (str):
                One of the supported monitoring objectives:
                ``raw-feature-drift`` ``prediction-output-drift``
                ``feature-attribution``
            model_monitoring_job (str):
                From a particular monitoring job.
            model_monitoring_schedule (str):
                From a particular monitoring schedule.
            algorithm (str):
                Specify the algorithm type used for distance calculation,
                eg: jensen_shannon_divergence, l_infinity.
        """

        stats_name: str = proto.Field(
            proto.STRING,
            number=1,
        )
        objective_type: str = proto.Field(
            proto.STRING,
            number=2,
        )
        model_monitoring_job: str = proto.Field(
            proto.STRING,
            number=3,
        )
        model_monitoring_schedule: str = proto.Field(
            proto.STRING,
            number=4,
        )
        algorithm: str = proto.Field(
            proto.STRING,
            number=5,
        )

    tabular_stats_filter: TabularStatsFilter = proto.Field(
        proto.MESSAGE,
        number=1,
        oneof="filter",
        message=TabularStatsFilter,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from grpc_interceptor.testing.protos import (
    dummy_pb2 as grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2,
)


class DummyServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Execute = channel.unary_unary(
            "/DummyService/Execute",
            request_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            response_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
        )
        self.ExecuteClientStream = channel.stream_unary(
            "/DummyService/ExecuteClientStream",
            request_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            response_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
        )
        self.ExecuteServerStream = channel.unary_stream(
            "/DummyService/ExecuteServerStream",
            request_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            response_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
        )
        self.ExecuteClientServerStream = channel.stream_stream(
            "/DummyService/ExecuteClientServerStream",
            request_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            response_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
        )


class DummyServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Execute(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExecuteClientStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExecuteServerStream(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")

    def ExecuteClientServerStream(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details("Method not implemented!")
        raise NotImplementedError("Method not implemented!")


def add_DummyServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
        "Execute": grpc.unary_unary_rpc_method_handler(
            servicer.Execute,
            request_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.FromString,
            response_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.SerializeToString,
        ),
        "ExecuteClientStream": grpc.stream_unary_rpc_method_handler(
            servicer.ExecuteClientStream,
            request_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.FromString,
            response_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.SerializeToString,
        ),
        "ExecuteServerStream": grpc.unary_stream_rpc_method_handler(
            servicer.ExecuteServerStream,
            request_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.FromString,
            response_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.SerializeToString,
        ),
        "ExecuteClientServerStream": grpc.stream_stream_rpc_method_handler(
            servicer.ExecuteClientServerStream,
            request_deserializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.FromString,
            response_serializer=grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.SerializeToString,
        ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
        "DummyService", rpc_method_handlers
    )
    server.add_generic_rpc_handlers((generic_handler,))


# This class is part of an EXPERIMENTAL API.
class DummyService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Execute(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_unary(
            request,
            target,
            "/DummyService/Execute",
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExecuteClientStream(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            "/DummyService/ExecuteClientStream",
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExecuteServerStream(
        request,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.unary_stream(
            request,
            target,
            "/DummyService/ExecuteServerStream",
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )

    @staticmethod
    def ExecuteClientServerStream(
        request_iterator,
        target,
        options=(),
        channel_credentials=None,
        call_credentials=None,
        insecure=False,
        compression=None,
        wait_for_ready=None,
        timeout=None,
        metadata=None,
    ):
        return grpc.experimental.stream_stream(
            request_iterator,
            target,
            "/DummyService/ExecuteClientServerStream",
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyRequest.SerializeToString,
            grpc__interceptor_dot_testing_dot_protos_dot_dummy__pb2.DummyResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
        )



# -*- coding: utf-8 -*-

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import re
from typing import Optional

from google.auth import credentials as auth_credentials
from google.cloud.aiplatform import base
from google.cloud.aiplatform import utils


class _PublisherModel(base.VertexAiResourceNoun):
    """Publisher Model Resource for Vertex AI."""

    client_class = utils.ModelGardenClientWithOverride

    _resource_noun = "publisher_model"
    _getter_method = "get_publisher_model"
    _delete_method = None
    _parse_resource_name_method = "parse_publisher_model_path"
    _format_resource_name_method = "publisher_model_path"

    def __init__(
        self,
        resource_name: str,
        project: Optional[str] = None,
        location: Optional[str] = None,
        credentials: Optional[auth_credentials.Credentials] = None,
    ):
        """Retrieves an existing PublisherModel resource given a resource name or model garden id.

        Args:
            resource_name (str):
                Required. A fully-qualified PublisherModel resource name or
                model garden id. Format:
                `publishers/{publisher}/models/{publisher_model}` or
                `{publisher}/{publisher_model}`.
            project (str):
                Optional. Project to retrieve the resource from. If not set,
                project set in aiplatform.init will be used.
            location (str):
                Optional. Location to retrieve the resource from. If not set,
                location set in aiplatform.init will be used.
            credentials (auth_credentials.Credentials):
                Optional. Custom credentials to use to retrieve the resource.
                Overrides credentials set in aiplatform.init.
        """

        super().__init__(project=project, location=location, credentials=credentials)

        if self._parse_resource_name(resource_name):
            full_resource_name = resource_name
        else:
            m = re.match(r"^(?P<publisher>.+?)/(?P<model>.+?)$", resource_name)
            if m:
                full_resource_name = self._format_resource_name(**m.groupdict())
            else:
                raise ValueError(
                    f"`{resource_name}` is not a valid PublisherModel resource "
                    "name or model garden id."
                )

        self._gca_resource = getattr(self.api_client, self._getter_method)(
            name=full_resource_name, retry=base._DEFAULT_RETRY
        )



# -*- coding: utf-8 -*-
# Copyright 2025 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from __future__ import annotations

from typing import MutableMapping, MutableSequence

import proto  # type: ignore

from google.cloud.aiplatform_v1.types import encryption_spec as gca_encryption_spec
from google.cloud.aiplatform_v1.types import machine_resources
from google.cloud.aiplatform_v1.types import service_networking
from google.protobuf import timestamp_pb2  # type: ignore
from google.rpc import status_pb2  # type: ignore


__protobuf__ = proto.module(
    package="google.cloud.aiplatform.v1",
    manifest={
        "PersistentResource",
        "ResourcePool",
        "ResourceRuntimeSpec",
        "RaySpec",
        "ResourceRuntime",
        "ServiceAccountSpec",
        "RayMetricSpec",
        "RayLogsSpec",
    },
)


class PersistentResource(proto.Message):
    r"""Represents long-lasting resources that are dedicated to users
    to runs custom workloads.
    A PersistentResource can have multiple node pools and each node
    pool can have its own machine spec.

    Attributes:
        name (str):
            Immutable. Resource name of a
            PersistentResource.
        display_name (str):
            Optional. The display name of the
            PersistentResource. The name can be up to 128
            characters long and can consist of any UTF-8
            characters.
        resource_pools (MutableSequence[google.cloud.aiplatform_v1.types.ResourcePool]):
            Required. The spec of the pools of different
            resources.
        state (google.cloud.aiplatform_v1.types.PersistentResource.State):
            Output only. The detailed state of a Study.
        error (google.rpc.status_pb2.Status):
            Output only. Only populated when persistent resource's state
            is ``STOPPING`` or ``ERROR``.
        create_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time when the PersistentResource
            was created.
        start_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time when the PersistentResource for the first
            time entered the ``RUNNING`` state.
        update_time (google.protobuf.timestamp_pb2.Timestamp):
            Output only. Time when the PersistentResource
            was most recently updated.
        labels (MutableMapping[str, str]):
            Optional. The labels with user-defined
            metadata to organize PersistentResource.

            Label keys and values can be no longer than 64
            characters (Unicode codepoints), can only
            contain lowercase letters, numeric characters,
            underscores and dashes. International characters
            are allowed.

            See https://goo.gl/xmQnxf for more information
            and examples of labels.
        network (str):
            Optional. The full name of the Compute Engine
            `network </compute/docs/networks-and-firewalls#networks>`__
            to peered with Vertex AI to host the persistent resources.
            For example, ``projects/12345/global/networks/myVPC``.
            `Format </compute/docs/reference/rest/v1/networks/insert>`__
            is of the form
            ``projects/{project}/global/networks/{network}``. Where
            {project} is a project number, as in ``12345``, and
            {network} is a network name.

            To specify this field, you must have already `configured VPC
            Network Peering for Vertex
            AI <https://cloud.google.com/vertex-ai/docs/general/vpc-peering>`__.

            If this field is left unspecified, the resources aren't
            peered with any network.
        psc_interface_config (google.cloud.aiplatform_v1.types.PscInterfaceConfig):
            Optional. Configuration for PSC-I for
            PersistentResource.
        encryption_spec (google.cloud.aiplatform_v1.types.EncryptionSpec):
            Optional. Customer-managed encryption key
            spec for a PersistentResource. If set, this
            PersistentResource and all sub-resources of this
            PersistentResource will be secured by this key.
        resource_runtime_spec (google.cloud.aiplatform_v1.types.ResourceRuntimeSpec):
            Optional. Persistent Resource runtime spec.
            For example, used for Ray cluster configuration.
        resource_runtime (google.cloud.aiplatform_v1.types.ResourceRuntime):
            Output only. Runtime information of the
            Persistent Resource.
        reserved_ip_ranges (MutableSequence[str]):
            Optional. A list of names for the reserved IP ranges under
            the VPC network that can be used for this persistent
            resource.

            If set, we will deploy the persistent resource within the
            provided IP ranges. Otherwise, the persistent resource is
            deployed to any IP ranges under the provided VPC network.

            Example: ['vertex-ai-ip-range'].
    """

    class State(proto.Enum):
        r"""Describes the PersistentResource state.

        Values:
            STATE_UNSPECIFIED (0):
                Not set.
            PROVISIONING (1):
                The PROVISIONING state indicates the
                persistent resources is being created.
            RUNNING (3):
                The RUNNING state indicates the persistent
                resource is healthy and fully usable.
            STOPPING (4):
                The STOPPING state indicates the persistent
                resource is being deleted.
            ERROR (5):
                The ERROR state indicates the persistent resource may be
                unusable. Details can be found in the ``error`` field.
            REBOOTING (6):
                The REBOOTING state indicates the persistent
                resource is being rebooted (PR is not available
                right now but is expected to be ready again
                later).
            UPDATING (7):
                The UPDATING state indicates the persistent
                resource is being updated.
        """

        STATE_UNSPECIFIED = 0
        PROVISIONING = 1
        RUNNING = 3
        STOPPING = 4
        ERROR = 5
        REBOOTING = 6
        UPDATING = 7

    name: str = proto.Field(
        proto.STRING,
        number=1,
    )
    display_name: str = proto.Field(
        proto.STRING,
        number=2,
    )
    resource_pools: MutableSequence["ResourcePool"] = proto.RepeatedField(
        proto.MESSAGE,
        number=4,
        message="ResourcePool",
    )
    state: State = proto.Field(
        proto.ENUM,
        number=5,
        enum=State,
    )
    error: status_pb2.Status = proto.Field(
        proto.MESSAGE,
        number=6,
        message=status_pb2.Status,
    )
    create_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=7,
        message=timestamp_pb2.Timestamp,
    )
    start_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=8,
        message=timestamp_pb2.Timestamp,
    )
    update_time: timestamp_pb2.Timestamp = proto.Field(
        proto.MESSAGE,
        number=9,
        message=timestamp_pb2.Timestamp,
    )
    labels: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=10,
    )
    network: str = proto.Field(
        proto.STRING,
        number=11,
    )
    psc_interface_config: service_networking.PscInterfaceConfig = proto.Field(
        proto.MESSAGE,
        number=17,
        message=service_networking.PscInterfaceConfig,
    )
    encryption_spec: gca_encryption_spec.EncryptionSpec = proto.Field(
        proto.MESSAGE,
        number=12,
        message=gca_encryption_spec.EncryptionSpec,
    )
    resource_runtime_spec: "ResourceRuntimeSpec" = proto.Field(
        proto.MESSAGE,
        number=13,
        message="ResourceRuntimeSpec",
    )
    resource_runtime: "ResourceRuntime" = proto.Field(
        proto.MESSAGE,
        number=14,
        message="ResourceRuntime",
    )
    reserved_ip_ranges: MutableSequence[str] = proto.RepeatedField(
        proto.STRING,
        number=15,
    )


class ResourcePool(proto.Message):
    r"""Represents the spec of a group of resources of the same type,
    for example machine type, disk, and accelerators, in a
    PersistentResource.


    .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

    Attributes:
        id (str):
            Immutable. The unique ID in a
            PersistentResource for referring to this
            resource pool. User can specify it if necessary.
            Otherwise, it's generated automatically.
        machine_spec (google.cloud.aiplatform_v1.types.MachineSpec):
            Required. Immutable. The specification of a
            single machine.
        replica_count (int):
            Optional. The total number of machines to use
            for this resource pool.

            This field is a member of `oneof`_ ``_replica_count``.
        disk_spec (google.cloud.aiplatform_v1.types.DiskSpec):
            Optional. Disk spec for the machine in this
            node pool.
        used_replica_count (int):
            Output only. The number of machines currently in use by
            training jobs for this resource pool. Will replace
            idle_replica_count.
        autoscaling_spec (google.cloud.aiplatform_v1.types.ResourcePool.AutoscalingSpec):
            Optional. Optional spec to configure GKE or
            Ray-on-Vertex autoscaling
    """

    class AutoscalingSpec(proto.Message):
        r"""The min/max number of replicas allowed if enabling
        autoscaling


        .. _oneof: https://proto-plus-python.readthedocs.io/en/stable/fields.html#oneofs-mutually-exclusive-fields

        Attributes:
            min_replica_count (int):
                Optional. min replicas in the node pool, must be ≤
                replica_count and < max_replica_count or will throw error.
                For autoscaling enabled Ray-on-Vertex, we allow
                min_replica_count of a resource_pool to be 0 to match the
                OSS Ray
                behavior(https://docs.ray.io/en/latest/cluster/vms/user-guides/configuring-autoscaling.html#cluster-config-parameters).
                As for Persistent Resource, the min_replica_count must be >
                0, we added a corresponding validation inside
                CreatePersistentResourceRequestValidator.java.

                This field is a member of `oneof`_ ``_min_replica_count``.
            max_replica_count (int):
                Optional. max replicas in the node pool, must be ≥
                replica_count and > min_replica_count or will throw error

                This field is a member of `oneof`_ ``_max_replica_count``.
        """

        min_replica_count: int = proto.Field(
            proto.INT64,
            number=1,
            optional=True,
        )
        max_replica_count: int = proto.Field(
            proto.INT64,
            number=2,
            optional=True,
        )

    id: str = proto.Field(
        proto.STRING,
        number=1,
    )
    machine_spec: machine_resources.MachineSpec = proto.Field(
        proto.MESSAGE,
        number=2,
        message=machine_resources.MachineSpec,
    )
    replica_count: int = proto.Field(
        proto.INT64,
        number=3,
        optional=True,
    )
    disk_spec: machine_resources.DiskSpec = proto.Field(
        proto.MESSAGE,
        number=4,
        message=machine_resources.DiskSpec,
    )
    used_replica_count: int = proto.Field(
        proto.INT64,
        number=6,
    )
    autoscaling_spec: AutoscalingSpec = proto.Field(
        proto.MESSAGE,
        number=7,
        message=AutoscalingSpec,
    )


class ResourceRuntimeSpec(proto.Message):
    r"""Configuration for the runtime on a PersistentResource instance,
    including but not limited to:

    - Service accounts used to run the workloads.
    - Whether to make it a dedicated Ray Cluster.

    Attributes:
        service_account_spec (google.cloud.aiplatform_v1.types.ServiceAccountSpec):
            Optional. Configure the use of workload
            identity on the PersistentResource
        ray_spec (google.cloud.aiplatform_v1.types.RaySpec):
            Optional. Ray cluster configuration.
            Required when creating a dedicated RayCluster on
            the PersistentResource.
    """

    service_account_spec: "ServiceAccountSpec" = proto.Field(
        proto.MESSAGE,
        number=2,
        message="ServiceAccountSpec",
    )
    ray_spec: "RaySpec" = proto.Field(
        proto.MESSAGE,
        number=1,
        message="RaySpec",
    )


class RaySpec(proto.Message):
    r"""Configuration information for the Ray cluster.
    For experimental launch, Ray cluster creation and Persistent
    cluster creation are 1:1 mapping: We will provision all the
    nodes within the Persistent cluster as Ray nodes.

    Attributes:
        image_uri (str):
            Optional. Default image for user to choose a preferred ML
            framework (for example, TensorFlow or Pytorch) by choosing
            from `Vertex prebuilt
            images <https://cloud.google.com/vertex-ai/docs/training/pre-built-containers>`__.
            Either this or the resource_pool_images is required. Use
            this field if you need all the resource pools to have the
            same Ray image. Otherwise, use the {@code
            resource_pool_images} field.
        resource_pool_images (MutableMapping[str, str]):
            Optional. Required if image_uri isn't set. A map of
            resource_pool_id to prebuild Ray image if user need to use
            different images for different head/worker pools. This map
            needs to cover all the resource pool ids. Example: {
            "ray_head_node_pool": "head image" "ray_worker_node_pool1":
            "worker image" "ray_worker_node_pool2": "another worker
            image" }
        head_node_resource_pool_id (str):
            Optional. This will be used to indicate which
            resource pool will serve as the Ray head
            node(the first node within that pool). Will use
            the machine from the first workerpool as the
            head node by default if this field isn't set.
        ray_metric_spec (google.cloud.aiplatform_v1.types.RayMetricSpec):
            Optional. Ray metrics configurations.
        ray_logs_spec (google.cloud.aiplatform_v1.types.RayLogsSpec):
            Optional. OSS Ray logging configurations.
    """

    image_uri: str = proto.Field(
        proto.STRING,
        number=1,
    )
    resource_pool_images: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=6,
    )
    head_node_resource_pool_id: str = proto.Field(
        proto.STRING,
        number=7,
    )
    ray_metric_spec: "RayMetricSpec" = proto.Field(
        proto.MESSAGE,
        number=8,
        message="RayMetricSpec",
    )
    ray_logs_spec: "RayLogsSpec" = proto.Field(
        proto.MESSAGE,
        number=10,
        message="RayLogsSpec",
    )


class ResourceRuntime(proto.Message):
    r"""Persistent Cluster runtime information as output

    Attributes:
        access_uris (MutableMapping[str, str]):
            Output only. URIs for user to connect to the Cluster.
            Example: { "RAY_HEAD_NODE_INTERNAL_IP": "head-node-IP:10001"
            "RAY_DASHBOARD_URI": "ray-dashboard-address:8888" }
    """

    access_uris: MutableMapping[str, str] = proto.MapField(
        proto.STRING,
        proto.STRING,
        number=1,
    )


class ServiceAccountSpec(proto.Message):
    r"""Configuration for the use of custom service account to run
    the workloads.

    Attributes:
        enable_custom_service_account (bool):
            Required. If true, custom user-managed service account is
            enforced to run any workloads (for example, Vertex Jobs) on
            the resource. Otherwise, uses the `Vertex AI Custom Code
            Service
            Agent <https://cloud.google.com/vertex-ai/docs/general/access-control#service-agents>`__.
        service_account (str):
            Optional. Required when all below conditions are met

            - ``enable_custom_service_account`` is true;
            - any runtime is specified via ``ResourceRuntimeSpec`` on
              creation time, for example, Ray

            The users must have ``iam.serviceAccounts.actAs`` permission
            on this service account and then the specified runtime
            containers will run as it.

            Do not set this field if you want to submit jobs using
            custom service account to this PersistentResource after
            creation, but only specify the ``service_account`` inside
            the job.
    """

    enable_custom_service_account: bool = proto.Field(
        proto.BOOL,
        number=1,
    )
    service_account: str = proto.Field(
        proto.STRING,
        number=2,
    )


class RayMetricSpec(proto.Message):
    r"""Configuration for the Ray metrics.

    Attributes:
        disabled (bool):
            Optional. Flag to disable the Ray metrics
            collection.
    """

    disabled: bool = proto.Field(
        proto.BOOL,
        number=1,
    )


class RayLogsSpec(proto.Message):
    r"""Configuration for the Ray OSS Logs.

    Attributes:
        disabled (bool):
            Optional. Flag to disable the export of Ray
            OSS logs to Cloud Logging.
    """

    disabled: bool = proto.Field(
        proto.BOOL,
        number=1,
    )


__all__ = tuple(sorted(__protobuf__.manifest))



# -*- coding: utf-8 -*-

# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses
from typing import Dict, List, Optional
from google.cloud.aiplatform_v1beta1.types import PersistentResource


@dataclasses.dataclass
class AutoscalingSpec:
    """Autoscaling spec for a ray cluster node.

    Attributes:
        min_replica_count: The minimum number of replicas in the cluster.
        max_replica_count: The maximum number of replicas in the cluster.
    """

    min_replica_count: int = 1
    max_replica_count: int = 2


@dataclasses.dataclass
class Resources:
    """Resources for a ray cluster node.

    Attributes:
        machine_type: See the list of machine types:
            https://cloud.google.com/vertex-ai/docs/training/configure-compute#machine-types
        node_count: This argument represents how many nodes to start for the
            ray cluster.
        accelerator_type: e.g. "NVIDIA_TESLA_P4".
            Vertex AI supports the following types of GPU:
            https://cloud.google.com/vertex-ai/docs/training/configure-compute#specifying_gpus
        accelerator_count: The number of accelerators to attach to the machine.
        boot_disk_type: Type of the boot disk (default is "pd-ssd").
            Valid values: "pd-ssd" (Persistent Disk Solid State Drive) or
            "pd-standard" (Persistent Disk Hard Disk Drive).
        boot_disk_size_gb: Size in GB of the boot disk (default is 100GB). Must
            be either unspecified or within the range of [100, 64000].
        custom_image: Custom image for this resource (e.g.
            us-docker.pkg.dev/my-project/ray-gpu.2-9.py310-tf:latest).
        autoscaling_spec: Autoscaling spec for this resource.
    """

    machine_type: Optional[str] = "n1-standard-16"
    node_count: Optional[int] = 1
    accelerator_type: Optional[str] = None
    accelerator_count: Optional[int] = 0
    boot_disk_type: Optional[str] = "pd-ssd"
    boot_disk_size_gb: Optional[int] = 100
    custom_image: Optional[str] = None
    autoscaling_spec: Optional[AutoscalingSpec] = None


@dataclasses.dataclass
class NodeImages:
    """Custom images for a ray cluster.

    We currently support Ray v2.9, v2.33, v2.42 and python v3.10.
    We also support python v3.11 for Ray v2.42 and v2.47.
    The custom images must be extended from the following base images:
    "{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.2-9.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.2-9.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.2-33.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.2-33.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.2-42.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.2-42.py310:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.2-42.py311:latest",
    "{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.2-42.py311:latest".
    "{region}-docker.pkg.dev/vertex-ai/training/ray-cpu.2-47.py311:latest", or
    "{region}-docker.pkg.dev/vertex-ai/training/ray-gpu.2-47.py311:latest". In
    order to use custom images, need to specify both head and worker images.

    Attributes:
        head: image for head node (eg. us-docker.pkg.dev/my-project/ray-cpu.2-47.py311:latest).
        worker: image for all worker nodes (eg. us-docker.pkg.dev/my-project/ray-gpu.2-47.py311:latest).
    """

    head: str = None
    worker: str = None


@dataclasses.dataclass
class PscIConfig:
    """PSC-I config.

    Attributes:
        network_attachment: Optional. The name or full name of the Compute Engine
            `network attachment <https://cloud.google.com/vpc/docs/about-network-attachments>`
            to attach to the resource. It has a format:
            ``projects/{project}/regions/{region}/networkAttachments/{networkAttachment}``.
            Where {project} is a project number, as in ``12345``, and
            {networkAttachment} is a network attachment name. To specify
            this field, you must have already [created a network
            attachment]
            (https://cloud.google.com/vpc/docs/create-manage-network-attachments#create-network-attachments).
            This field is only used for resources using PSC-I. Make sure you do not
            specify the network here for VPC peering.
    """

    network_attachment: str = None


@dataclasses.dataclass
class NfsMount:
    """NFS mount.

    Attributes:
        server: Required. IP address of the NFS server.
        path: Required. Source path exported from NFS server. Has to start
            with '/', and combined with the ip address, it indicates the
            source mount path in the form of ``server:path``.
        mount_point: Required. Destination mount path. The NFS will be mounted
            for the user under /mnt/nfs/<mount_point>.
    """

    server: str = None
    path: str = None
    mount_point: str = None


@dataclasses.dataclass
class Cluster:
    """Ray cluster (output only).

    Attributes:
        cluster_resource_name: It has a format:
            "projects/<project_num>/locations/<region>/persistentResources/<pr_id>".
        network: Virtual private cloud (VPC) network. It has a format:
            "projects/<project_num>/global/networks/<network_name>".
            For Ray Client, VPC peering is required to connect to the cluster
            managed in the Vertex API service. For Ray Job API, VPC network is
            not required because cluster connection can be accessed through
            dashboard address.
        reserved_ip_ranges: A list of names for the reserved IP ranges under
            the VPC network that can be used for this cluster. If set, we will
            deploy the cluster within the provided IP ranges. Otherwise, the
            cluster is deployed to any IP ranges under the provided VPC network.
            Example: ["vertex-ai-ip-range"].
        service_account: Service account to be used for running Ray programs on
            the cluster.
        state: Describes the cluster state (defined in PersistentResource.State).
        python_version: Python version for the ray cluster (e.g. "3.11").
        ray_version: Ray version for the ray cluster (e.g. "2.47").
        head_node_type: The head node resource. Resources.node_count must be 1.
            If not set, by default it is a CPU node with machine_type of n1-standard-8.
        worker_node_types: The list of Resources of the worker nodes. Should not
            duplicate the elements in the list.
        dashboard_address: For Ray Job API (JobSubmissionClient), with this
           cluster connection doesn't require VPC peering.
        labels:
            The labels with user-defined metadata to organize Ray cluster.

            Label keys and values can be no longer than 64 characters (Unicode
            codepoints), can only contain lowercase letters, numeric characters,
            underscores and dashes. International characters are allowed.

            See https://goo.gl/xmQnxf for more information and examples of labels.
    """

    cluster_resource_name: str = None
    network: str = None
    reserved_ip_ranges: List[str] = None
    service_account: str = None
    state: PersistentResource.State = None
    python_version: str = None
    ray_version: str = None
    head_node_type: Resources = None
    worker_node_types: List[Resources] = None
    dashboard_address: str = None
    ray_metric_enabled: bool = True
    ray_logs_enabled: bool = True
    psc_interface_config: PscIConfig = None
    labels: Dict[str, str] = None


def _check_machine_spec_identical(
    node_type_1: Resources,
    node_type_2: Resources,
) -> int:
    """Check if node_type_1 and node_type_2 have the same machine_spec.
    If they are identical, return additional_replica_count."""
    additional_replica_count = 0

    # Check if machine_spec are the same
    if (
        node_type_1.machine_type == node_type_2.machine_type
        and node_type_1.accelerator_type == node_type_2.accelerator_type
        and node_type_1.accelerator_count == node_type_2.accelerator_count
    ):
        if node_type_1.boot_disk_type != node_type_2.boot_disk_type:
            raise ValueError(
                "Worker disk type must match the head node's disk type if"
                " sharing the same machine_type, accelerator_type, and"
                " accelerator_count"
            )
        if node_type_1.boot_disk_size_gb != node_type_2.boot_disk_size_gb:
            raise ValueError(
                "Worker disk size must match the head node's disk size if"
                " sharing the same machine_type, accelerator_type, and"
                " accelerator_count"
            )
        additional_replica_count = node_type_2.node_count
        return additional_replica_count

    return additional_replica_count



# -*- coding: utf-8 -*-
# Copyright 2023 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import dataclasses
import inspect
import json
import os
import sys
import types
import typing
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    TypedDict,
    Union,
)
from importlib import metadata as importlib_metadata

import proto

from google.cloud.aiplatform import base
from google.api import httpbody_pb2
from google.protobuf import struct_pb2
from google.protobuf import json_format

try:
    # For LangChain templates, they might not import langchain_core and get
    #   PydanticUserError: `query` is not fully defined; you should define
    #   `RunnableConfig`, then call `query.model_rebuild()`.
    import langchain_core.runnables.config

    RunnableConfig = langchain_core.runnables.config.RunnableConfig
except ImportError:
    RunnableConfig = Any

try:
    import packaging

    SpecifierSet = packaging.specifiers.SpecifierSet
except AttributeError:
    SpecifierSet = Any

try:
    _BUILTIN_MODULE_NAMES: Sequence[str] = sys.builtin_module_names
except AttributeError:
    _BUILTIN_MODULE_NAMES: Sequence[str] = []

try:
    # sys.stdlib_module_names is available from Python 3.10 onwards.
    _STDLIB_MODULE_NAMES: frozenset = sys.stdlib_module_names
except AttributeError:
    _STDLIB_MODULE_NAMES: frozenset = frozenset()

try:
    _PACKAGE_DISTRIBUTIONS: Mapping[str, Sequence[str]] = (
        importlib_metadata.packages_distributions()
    )

except AttributeError:
    _PACKAGE_DISTRIBUTIONS: Mapping[str, Sequence[str]] = {}

try:
    from autogen.agentchat import chat

    AutogenChatResult = chat.ChatResult
except ImportError:
    AutogenChatResult = Any

try:
    from autogen.io import run_response

    AutogenRunResponse = run_response.RunResponse
except ImportError:
    AutogenRunResponse = Any

try:
    import pydantic

    BaseModel = pydantic.BaseModel
except ImportError:
    BaseModel = Any

JsonDict = Dict[str, Any]


class _RequirementsValidationActions(TypedDict):
    append: Set[str]


class _RequirementsValidationWarnings(TypedDict):
    missing: Set[str]
    incompatible: Set[str]


class _RequirementsValidationResult(TypedDict):
    warnings: _RequirementsValidationWarnings
    actions: _RequirementsValidationActions


LOGGER = base.Logger("vertexai.agent_engines")

_BASE_MODULES = set(_BUILTIN_MODULE_NAMES + tuple(_STDLIB_MODULE_NAMES))
_DEFAULT_REQUIRED_PACKAGES = frozenset(["cloudpickle", "pydantic"])
_ACTIONS_KEY = "actions"
_ACTION_APPEND = "append"
_WARNINGS_KEY = "warnings"
_WARNING_MISSING = "missing"
_WARNING_INCOMPATIBLE = "incompatible"
_INSTALLATION_SUBDIR = "installation_scripts"


def to_proto(
    obj: Union[JsonDict, proto.Message],
    message: Optional[proto.Message] = None,
) -> proto.Message:
    """Parses a JSON-like object into a message.

    If the object is already a message, this will return the object as-is. If
    the object is a JSON Dict, this will parse and merge the object into the
    message.

    Args:
        obj (Union[dict[str, Any], proto.Message]):
            Required. The object to convert to a proto message.
        message (proto.Message):
            Optional. A protocol buffer message to merge the obj into. It
            defaults to Struct() if unspecified.

    Returns:
        proto.Message: The same message passed as argument.
    """
    if message is None:
        message = struct_pb2.Struct()
    if isinstance(obj, (proto.Message, struct_pb2.Struct)):
        return obj
    try:
        json_format.ParseDict(obj, message._pb)
    except AttributeError:
        json_format.ParseDict(obj, message)
    return message


def to_dict(message: proto.Message) -> JsonDict:
    """Converts the contents of the protobuf message to JSON format.

    Args:
        message (proto.Message):
            Required. The proto message to be converted to a JSON dictionary.

    Returns:
        dict[str, Any]: A dictionary containing the contents of the proto.
    """
    try:
        # Best effort attempt to convert the message into a JSON dictionary.
        result: JsonDict = json.loads(
            json_format.MessageToJson(
                message._pb,
                preserving_proto_field_name=True,
            )
        )
    except AttributeError:
        result: JsonDict = json.loads(
            json_format.MessageToJson(
                message,
                preserving_proto_field_name=True,
            )
        )
    return result


def _dataclass_to_dict_or_raise(obj: Any) -> JsonDict:
    """Converts a dataclass to a JSON dictionary.

    Args:
        obj (Any):
            Required. The dataclass to be converted to a JSON dictionary.

    Returns:
        dict[str, Any]: A dictionary containing the contents of the dataclass.

    Raises:
        TypeError: If the object is not a dataclass.
    """
    if not dataclasses.is_dataclass(obj):
        raise TypeError(f"Object is not a dataclass: {obj}")
    return json.loads(json.dumps(dataclasses.asdict(obj)))


def _autogen_run_response_protocol_to_dict(
    obj: AutogenRunResponse,
) -> JsonDict:
    """Converts an AutogenRunResponse object into a JSON-serializable dictionary.

    This function takes a `RunResponseProtocol` object and transforms its
    relevant attributes into a dictionary format suitable for JSON conversion.

    The `RunResponseProtocol` defines the structure of the response object,
    which typically includes:

    *   **summary** (`Optional[str]`):
        A textual summary of the run.
    *   **messages** (`Iterable[Message]`):
        A sequence of messages exchanged during the run.
        Each message is expected to be a JSON-serializable dictionary (`Dict[str,
        Any]`).
    *   **events** (`Iterable[BaseEvent]`):
        A sequence of events that occurred during the run.
        Note: The `process()` method, if present, is called before conversion,
        which typically clears this event queue.
    *   **context_variables** (`Optional[dict[str, Any]]`):
        A dictionary containing contextual variables from the run.
    *   **last_speaker** (`Optional[Agent]`):
        The agent that produced the last message.
        The `Agent` object has attributes like `name` (Optional[str]) and
        `description` (Optional[str]).
    *   **cost** (`Optional[Cost]`):
        Information about the computational cost of the run.
        The `Cost` object inherits from `pydantic.BaseModel` and is converted
        to JSON using its `model_dump_json()` method.
    *   **process** (`Optional[Callable[[], None]]`):
        An optional function (like a console event processor) that is called
        before the conversion takes place.
        Executing this method often clears the `events` queue.

    For a detailed definition of `RunResponseProtocol` and its components, refer
    to: https://github.com/ag2ai/ag2/blob/main/autogen/io/run_response.py

    Args:
        obj (AutogenRunResponse): The AutogenRunResponse object to convert. This
            object must conform to the `RunResponseProtocol`.

    Returns:
        JsonDict: A dictionary representation of the AutogenRunResponse, ready
            to be serialized into JSON. The dictionary includes keys like
            'summary', 'messages', 'context_variables', 'last_speaker_name',
            and 'cost'.
    """
    if hasattr(obj, "process"):
        obj.process()

    last_speaker = None
    if getattr(obj, "last_speaker", None) is not None:
        last_speaker = {
            "name": getattr(obj.last_speaker, "name", None),
            "description": getattr(obj.last_speaker, "description", None),
        }

    cost = None
    if getattr(obj, "cost", None) is not None:
        if hasattr(obj.cost, "model_dump_json"):
            cost = json.loads(obj.cost.model_dump_json())
        else:
            cost = str(obj.cost)

    result = {
        "summary": getattr(obj, "summary", None),
        "messages": list(getattr(obj, "messages", [])),
        "context_variables": getattr(obj, "context_variables", None),
        "last_speaker": last_speaker,
        "cost": cost,
    }
    return json.loads(json.dumps(result))


def to_json_serializable_autogen_object(
    obj: Union[
        AutogenChatResult,
        AutogenRunResponse,
    ],
) -> JsonDict:
    """Converts an Autogen object to a JSON serializable object.

    In `ag2<=0.8.4`, `.run()` will return a `ChatResult` object.
    In `ag2>=0.8.5`, `.run()` will return a `RunResponse` object.

    Args:
        obj (Union[AutogenChatResult, AutogenRunResponse]):
            Required. The Autogen object to be converted to a JSON serializable
            object.

    Returns:
        JsonDict: A JSON serializable object.
    """
    if isinstance(obj, AutogenChatResult):
        return _dataclass_to_dict_or_raise(obj)
    return _autogen_run_response_protocol_to_dict(obj)


def yield_parsed_json(body: httpbody_pb2.HttpBody) -> Iterable[Any]:
    """Converts the contents of the httpbody message to JSON format.

    Args:
        body (httpbody_pb2.HttpBody):
            Required. The httpbody body to be converted to a JSON.

    Yields:
        Any: A JSON object or the original body if it is not JSON or None.
    """
    content_type = getattr(body, "content_type", None)
    data = getattr(body, "data", None)

    if content_type is None or data is None or "application/json" not in content_type:
        yield body
        return

    try:
        utf8_data = data.decode("utf-8")
    except Exception as e:
        LOGGER.warning(f"Failed to decode data: {data}. Exception: {e}")
        yield body
        return

    if not utf8_data:
        yield None
        return

    # Handle the case of multiple dictionaries delimited by newlines.
    for line in utf8_data.split("\n"):
        if line:
            try:
                line = json.loads(line)
            except Exception as e:
                LOGGER.warning(f"failed to parse json: {line}. Exception: {e}")
            yield line


def parse_constraints(
    constraints: Sequence[str],
) -> Mapping[str, "SpecifierSet"]:
    """Parses a list of constraints into a dict of requirements.

    Args:
        constraints (list[str]):
            Required. The list of package requirements to parse. This is assumed
            to come from the `requirements.txt` file.

    Returns:
        dict[str, SpecifierSet]: The specifiers for each package.
    """
    requirements = _import_packaging_requirements_or_raise()
    result = {}
    for constraint in constraints:
        try:
            if constraint.endswith(".whl"):
                constraint = os.path.basename(constraint)
            requirement = requirements.Requirement(constraint)
        except Exception as e:
            LOGGER.warning(f"Failed to parse constraint: {constraint}. Exception: {e}")
            continue
        result[requirement.name] = requirement.specifier or None
    return result


def validate_requirements_or_warn(
    obj: Any,
    requirements: List[str],
    logger: base.Logger = LOGGER,
) -> Mapping[str, str]:
    """Compiles the requirements into a list of requirements."""
    requirements = requirements.copy()
    try:
        current_requirements = scan_requirements(obj)
        logger.info(f"Identified the following requirements: {current_requirements}")
        constraints = parse_constraints(requirements)
        missing_requirements = compare_requirements(current_requirements, constraints)
        for warning_type, warnings in missing_requirements.get(
            _WARNINGS_KEY, {}
        ).items():
            if warnings:
                logger.warning(
                    f"The following requirements are {warning_type}: {warnings}"
                )
        for action_type, actions in missing_requirements.get(_ACTIONS_KEY, {}).items():
            if actions and action_type == _ACTION_APPEND:
                for action in actions:
                    requirements.append(action)
                logger.info(f"The following requirements are appended: {actions}")
    except Exception as e:
        logger.warning(f"Failed to compile requirements: {e}")
    return requirements


def compare_requirements(
    requirements: Mapping[str, str],
    constraints: Union[Sequence[str], Mapping[str, "SpecifierSet"]],
    *,
    required_packages: Optional[Sequence[str]] = None,
) -> Mapping[str, Mapping[str, Any]]:
    """Compares the requirements with the constraints.

    Args:
        requirements (Mapping[str, str]):
            Required. The packages (and their versions) to compare with the constraints.
            This is assumed to be the result of `scan_requirements`.
        constraints (Union[Sequence[str], Mapping[str, SpecifierSet]]):
            Required. The package constraints to compare against. This is assumed
            to be the result of `parse_constraints`.
        required_packages (Sequence[str]):
            Optional. The set of packages that are required to be in the
            constraints. It defaults to the set of packages that are required
            for deployment on Agent Engine.

    Returns:
        dict[str, dict[str, Any]]: The comparison result as a dictionary containing:
            * warnings:
                * missing: The set of packages that are not in the constraints.
                * incompatible: The set of packages that are in the constraints
                    but have versions that are not in the constraint specifier.
            * actions:
                * append: The set of packages that are not in the constraints
                    but should be appended to the constraints.
    """
    packaging_version = _import_packaging_version_or_raise()
    if required_packages is None:
        required_packages = _DEFAULT_REQUIRED_PACKAGES
    result = _RequirementsValidationResult(
        warnings=_RequirementsValidationWarnings(missing=set(), incompatible=set()),
        actions=_RequirementsValidationActions(append=set()),
    )
    if isinstance(constraints, list):
        constraints = parse_constraints(constraints)
    for package, package_version in requirements.items():
        if package not in constraints:
            result[_WARNINGS_KEY][_WARNING_MISSING].add(package)
            if package in required_packages:
                result[_ACTIONS_KEY][_ACTION_APPEND].add(
                    f"{package}=={package_version}"
                )
            continue
        if package_version:
            package_specifier = constraints[package]
            if not package_specifier:
                continue
            if packaging_version.Version(package_version) not in package_specifier:
                result[_WARNINGS_KEY][_WARNING_INCOMPATIBLE].add(
                    f"{package}=={package_version} (required: {str(package_specifier)})"
                )
    return result


def scan_requirements(
    obj: Any,
    ignore_modules: Optional[Sequence[str]] = None,
    package_distributions: Optional[Mapping[str, Sequence[str]]] = None,
    inspect_getmembers_kwargs: Optional[Mapping[str, Any]] = None,
) -> Mapping[str, str]:
    """Scans the object for modules and returns the requirements discovered.

    This is not a comprehensive scan of the object, and only detects for common
    cases based on the members of the object returned by `dir(obj)`.

    Args:
        obj (Any):
            Required. The object to scan for package requirements.
        ignore_modules (Sequence[str]):
            Optional. The set of modules to ignore. It defaults to the set of
            built-in and stdlib modules.
        package_distributions (Mapping[str, Sequence[str]]):
            Optional. The mapping of module names to the set of packages that
            contain them. It defaults to the set of packages from
            `importlib_metadata.packages_distributions()`.
        inspect_getmembers_kwargs (Mapping[str, Any]):
            Optional. The keyword arguments to pass to `inspect.getmembers`. It
            defaults to an empty dictionary.

    Returns:
        Sequence[str]: The list of requirements that were discovered.
    """
    if ignore_modules is None:
        ignore_modules = _BASE_MODULES
    if package_distributions is None:
        package_distributions = _PACKAGE_DISTRIBUTIONS
    modules_found = set(_DEFAULT_REQUIRED_PACKAGES)
    inspect_getmembers_kwargs = inspect_getmembers_kwargs or {}
    for _, attr in inspect.getmembers(obj, **inspect_getmembers_kwargs):
        if not attr or inspect.isbuiltin(attr) or not hasattr(attr, "__module__"):
            continue
        module_name = (attr.__module__ or "").split(".")[0]
        if module_name and module_name not in ignore_modules:
            for module in package_distributions.get(module_name, []):
                modules_found.add(module)
    return {module: importlib_metadata.version(module) for module in modules_found}


def _is_pydantic_serializable(param: inspect.Parameter) -> bool:
    """Checks if the parameter is pydantic serializable."""

    if param.annotation == inspect.Parameter.empty:
        return True

    if isinstance(param.annotation, str):
        return False
    pydantic = _import_pydantic_or_raise()
    try:
        pydantic.TypeAdapter(param.annotation)
        return True
    except Exception:
        return False


def generate_schema(
    f: Callable[..., Any],
    *,
    schema_name: Optional[str] = None,
    descriptions: Mapping[str, str] = {},
    required: Sequence[str] = [],
) -> JsonDict:
    """Generates the OpenAPI Schema for a callable object.

    Only positional and keyword arguments of the function `f` will be supported
    in the OpenAPI Schema that is generated. I.e. `*args` and `**kwargs` will
    not be present in the OpenAPI schema returned from this function. For those
    cases, you can either include it in the docstring for `f`, or modify the
    OpenAPI schema returned from this function to include additional arguments.

    Args:
        f (Callable):
            Required. The function to generate an OpenAPI Schema for.
        schema_name (str):
            Optional. The name for the OpenAPI schema. If unspecified, the name
            of the Callable will be used.
        descriptions (Mapping[str, str]):
            Optional. A `{name: description}` mapping for annotating input
            arguments of the function with user-provided descriptions. It
            defaults to an empty dictionary (i.e. there will not be any
            description for any of the inputs).
        required (Sequence[str]):
            Optional. For the user to specify the set of required arguments in
            function calls to `f`. If specified, it will be automatically
            inferred from `f`.

    Returns:
        dict[str, Any]: The OpenAPI Schema for the function `f` in JSON format.
    """
    pydantic = _import_pydantic_or_raise()
    defaults = dict(inspect.signature(f).parameters)
    fields_dict = {
        name: (
            # 1. We infer the argument type here: use Any rather than None so
            # it will not try to auto-infer the type based on the default value.
            (param.annotation if param.annotation != inspect.Parameter.empty else Any),
            pydantic.Field(
                # 2. We do not support default values for now.
                # default=(
                #     param.default if param.default != inspect.Parameter.empty
                #     else None
                # ),
                # 3. We support user-provided descriptions.
                description=descriptions.get(name, None),
            ),
        )
        for name, param in defaults.items()
        # We do not support *args or **kwargs
        if param.kind
        in (
            inspect.Parameter.POSITIONAL_OR_KEYWORD,
            inspect.Parameter.KEYWORD_ONLY,
            inspect.Parameter.POSITIONAL_ONLY,
        )
        and _is_pydantic_serializable(param)
    }
    parameters = pydantic.create_model(f.__name__, **fields_dict).schema()
    # Postprocessing
    # 4. Suppress unnecessary title generation:
    #    * https://github.com/pydantic/pydantic/issues/1051
    #    * http://cl/586221780
    parameters.pop("title", "")
    for name, function_arg in parameters.get("properties", {}).items():
        function_arg.pop("title", "")
        annotation = defaults[name].annotation
        # 5. Nullable fields:
        #     * https://github.com/pydantic/pydantic/issues/1270
        #     * https://stackoverflow.com/a/58841311
        #     * https://github.com/pydantic/pydantic/discussions/4872
        if typing.get_origin(annotation) is Union and type(None) in typing.get_args(
            annotation
        ):
            # for "typing.Optional" arguments, function_arg might be a
            # dictionary like
            #
            #   {'anyOf': [{'type': 'integer'}, {'type': 'null'}]
            for schema in function_arg.pop("anyOf", []):
                schema_type = schema.get("type")
                if schema_type and schema_type != "null":
                    function_arg["type"] = schema_type
                    break
            function_arg["nullable"] = True
    # 6. Annotate required fields.
    if required:
        # We use the user-provided "required" fields if specified.
        parameters["required"] = required
    else:
        # Otherwise we infer it from the function signature.
        parameters["required"] = [
            k
            for k in defaults
            if (
                defaults[k].default == inspect.Parameter.empty
                and defaults[k].kind
                in (
                    inspect.Parameter.POSITIONAL_OR_KEYWORD,
                    inspect.Parameter.KEYWORD_ONLY,
                    inspect.Parameter.POSITIONAL_ONLY,
                )
            )
        ]
    schema = dict(name=f.__name__, description=f.__doc__, parameters=parameters)
    if schema_name:
        schema["name"] = schema_name
    return schema


def is_noop_or_proxy_tracer_provider(tracer_provider) -> bool:
    """Returns True if the tracer_provider is Proxy or NoOp."""
    opentelemetry = _import_opentelemetry_or_warn()
    ProxyTracerProvider = opentelemetry.trace.ProxyTracerProvider
    NoOpTracerProvider = opentelemetry.trace.NoOpTracerProvider
    return isinstance(tracer_provider, (NoOpTracerProvider, ProxyTracerProvider))


def dump_event_for_json(event: BaseModel) -> Dict[str, Any]:
    """Dumps an ADK event to a JSON-serializable dictionary."""
    return json.loads(event.model_dump_json(exclude_none=True))


def _import_cloud_storage_or_raise() -> types.ModuleType:
    """Tries to import the Cloud Storage module."""
    try:
        from google.cloud import storage
    except ImportError as e:
        raise ImportError(
            "Cloud Storage is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        ) from e
    return storage


def _import_cloudpickle_or_raise() -> types.ModuleType:
    """Tries to import the cloudpickle module."""
    try:
        import cloudpickle  # noqa:F401
    except ImportError as e:
        raise ImportError(
            "cloudpickle is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        ) from e
    return cloudpickle


def _import_pydantic_or_raise() -> types.ModuleType:
    """Tries to import the pydantic module."""
    try:
        import pydantic

        _ = pydantic.Field
    except AttributeError:
        from pydantic import v1 as pydantic
    except ImportError as e:
        raise ImportError(
            "pydantic is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        ) from e
    return pydantic


def _import_packaging_requirements_or_raise() -> types.ModuleType:
    """Tries to import the packaging.requirements module."""
    try:
        from packaging import requirements
    except ImportError as e:
        raise ImportError(
            "packaging.requirements is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        ) from e
    return requirements


def _import_packaging_version_or_raise() -> types.ModuleType:
    """Tries to import the packaging.requirements module."""
    try:
        from packaging import version
    except ImportError as e:
        raise ImportError(
            "packaging.version is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        ) from e
    return version


def _import_opentelemetry_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the opentelemetry module."""
    try:
        import opentelemetry  # noqa:F401

        return opentelemetry
    except ImportError:
        LOGGER.warning(
            "opentelemetry-sdk is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        )
    return None


def _import_opentelemetry_sdk_trace_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the opentelemetry.sdk.trace module."""
    try:
        import opentelemetry.sdk.trace  # noqa:F401

        return opentelemetry.sdk.trace
    except ImportError:
        LOGGER.warning(
            "opentelemetry-sdk is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        )
    return None


def _import_cloud_trace_v2_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the google.cloud.trace_v2 module."""
    try:
        import google.cloud.trace_v2

        return google.cloud.trace_v2
    except ImportError:
        LOGGER.warning(
            "google-cloud-trace is not installed. Please call "
            "'pip install google-cloud-aiplatform[agent_engines]'."
        )
    return None


def _import_cloud_trace_exporter_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the opentelemetry.exporter.cloud_trace module."""
    try:
        import opentelemetry.exporter.cloud_trace  # noqa:F401

        return opentelemetry.exporter.cloud_trace
    except ImportError:
        LOGGER.warning(
            "opentelemetry-exporter-gcp-trace is not installed. Please "
            "call 'pip install google-cloud-aiplatform[agent_engines]'."
        )
    return None


def _import_openinference_langchain_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the openinference.instrumentation.langchain module."""
    try:
        import openinference.instrumentation.langchain  # noqa:F401

        return openinference.instrumentation.langchain
    except ImportError:
        LOGGER.warning(
            "openinference-instrumentation-langchain is not installed. Please "
            "call 'pip install google-cloud-aiplatform[langchain]'."
        )
    return None


def _import_openinference_autogen_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the openinference.instrumentation.autogen module."""
    try:
        import openinference.instrumentation.autogen  # noqa:F401

        return openinference.instrumentation.autogen
    except ImportError:
        LOGGER.warning(
            "openinference-instrumentation-autogen is not installed. Please "
            "call 'pip install google-cloud-aiplatform[ag2]'."
        )
    return None


def _import_autogen_tools_or_warn() -> Optional[types.ModuleType]:
    """Tries to import the autogen.tools module."""
    try:
        from autogen import tools

        return tools
    except ImportError:
        LOGGER.warning(
            "autogen.tools is not installed. Please "
            "call `pip install google-cloud-aiplatform[ag2]`."
        )
    return None


def validate_installation_scripts_or_raise(
    script_paths: Sequence[str],
    extra_packages: Sequence[str],
):
    """Validates the installation scripts' path explicitly provided by the user.

    Args:
        script_paths (Sequence[str]):
            Required. The paths to the installation scripts.
        extra_packages (Sequence[str]):
            Required. The extra packages to be updated.

    Raises:
        ValueError: If a user-defined script is not under the expected
            subdirectory, or not in `extra_packages`, or if an extra package is
            in the installation scripts subdirectory, but is not specified as an
            installation script.
    """
    for script_path in script_paths:
        if not script_path.startswith(_INSTALLATION_SUBDIR):
            LOGGER.warning(
                f"User-defined installation script '{script_path}' is not in "
                f"the expected '{_INSTALLATION_SUBDIR}' subdirectory. "
                f"Ensure it is placed in '{_INSTALLATION_SUBDIR}' within your "
                f"`extra_packages`."
            )
            raise ValueError(
                f"Required installation script '{script_path}' "
                f"is not under '{_INSTALLATION_SUBDIR}'"
            )

        if script_path not in extra_packages:
            LOGGER.warning(
                f"User-defined installation script '{script_path}' is not in "
                f"extra_packages. Ensure it is added to `extra_packages`."
            )
            raise ValueError(
                f"User-defined installation script '{script_path}' "
                f"does not exist in `extra_packages`"
            )

    for extra_package in extra_packages:
        if (
            extra_package.startswith(_INSTALLATION_SUBDIR)
            and extra_package not in script_paths
        ):
            LOGGER.warning(
                f"Extra package '{extra_package}' is in the installation "
                "scripts subdirectory, but is not specified as an installation "
                "script in `build_options`. "
                "Ensure it is added to installation_scripts for "
                "automatic execution."
            )
            raise ValueError(
                f"Extra package '{extra_package}' is in the installation "
                "scripts subdirectory, but is not specified as an installation "
                "script in `build_options`."
            )
    return


def _validate_resource_limits_or_raise(resource_limits: dict[str, str]) -> None:
    """Validates the resource limits.

    Checks that the resource limits are a dict with 'cpu' and 'memory' keys.
    Checks that the 'cpu' value is one of 1, 2, 4, 6, 8.
    Checks that the 'memory' value is a string ending with 'Gi'.
    Checks that the memory size is smaller than 32Gi.
    Checks that the memory size requires at least the specified number of CPUs.

    Args:
        resource_limits: The resource limits to be validated.

    Raises:
        TypeError: If the resource limits are not a dict.
        KeyError: If the resource limits do not contain 'cpu' and 'memory' keys.
        ValueError: If the 'cpu' value is not one of 1, 2, 4, 6, 8.
        ValueError: If the 'memory' value is not a string ending with 'Gi'.
        ValueError: If the memory size is too large.
        ValueError: If the memory size requires more CPUs than the specified
        'cpu' value.
    """
    if not isinstance(resource_limits, dict):
        raise TypeError(f"resource_limits must be a dict. Got {type(resource_limits)}")
    if "cpu" not in resource_limits or "memory" not in resource_limits:
        raise KeyError("resource_limits must contain 'cpu' and 'memory' keys.")

    cpu = int(resource_limits["cpu"])
    memory_str = resource_limits["memory"]

    if cpu not in [1, 2, 4, 6, 8]:
        raise ValueError(
            "resource_limits['cpu'] must be one of 1, 2, 4, 6, 8. Got" f" {cpu}"
        )

    if not isinstance(memory_str, str) or not memory_str.endswith("Gi"):
        raise ValueError(
            "resource_limits['memory'] must be a string ending with 'Gi'."
            f" Got {memory_str}"
        )

    try:
        memory_gb = int(memory_str[:-2])
    except ValueError:
        raise ValueError(
            f"Invalid memory value: {memory_str}. Must be an integer"
            " followed by 'Gi'."
        )

    # https://cloud.google.com/run/docs/configuring/memory-limits
    if memory_gb > 32:
        raise ValueError(
            f"Memory size of {memory_str} is too large. Must be smaller than 32Gi."
        )
    if memory_gb > 24:
        min_cpu = 8
    elif memory_gb > 16:
        min_cpu = 6
    elif memory_gb > 8:
        min_cpu = 4
    elif memory_gb > 4:
        min_cpu = 2
    else:
        min_cpu = 1

    if cpu < min_cpu:
        raise ValueError(
            f"Memory size of {memory_str} requires at least {min_cpu} CPUs."
            f" Got {cpu}"
        )
